"""
Extension for the `CommandCog`.
Loading this file via `commands.Bot.load_extension` will add `CommandCog` to the bot.
"""

import logging
import validators
from typing import Tuple

import aiohttp
import textwrap
import feedparser
from markdownify import markdownify
from discord import Interaction, Embed, Colour
from discord.ext import commands
from discord.app_commands import Choice, Group, autocomplete, choices
from sqlalchemy import insert, select, and_, delete

from feed import get_source, Source
from db import DatabaseManager, SentArticleModel, RssSourceModel

log = logging.getLogger(__name__)

rss_list_sort_choices = [
    Choice(name="Nickname", value=0),
    Choice(name="Date Added", value=1)
]

async def get_rss_data(url: str):
    async with aiohttp.ClientSession() as session:
        async with session.get(url) as response:
            items = await response.text(), response.status

    return items

async def followup(inter: Interaction, *args, **kwargs):
    """Shorthand for following up on an interaction.

    Parameters
    ----------
    inter : Interaction
        Represents an app command interaction.
    """

    await inter.followup.send(*args, **kwargs)

async def audit(cog, *args, **kwargs):
    """Shorthand for auditing an interaction."""

    await cog.bot.audit(*args, **kwargs)

# TODO SECURITY: a potential attack is that the user submits an rss feed then changes the target resource.
# Run a period task to check this.
async def validate_rss_source(nickname: str, url: str) -> Tuple[str | None, feedparser.FeedParserDict | None]:
    """Validate a provided RSS source.

    Parameters
    ----------
    nickname : str
        Nickname of the source. Must not contain URL.
    url : str
        URL of the source. Must be URL with valid status code and be an RSS feed.

    Returns
    -------
    str or None
        String invalid message if invalid, NoneType if valid.
    FeedParserDict or None
        The feed parsed from the given URL or None if invalid.
    """

    # Ensure the URL is valid
    if not validators.url(url):
        return f"The URL you have entered is malformed or invalid:\n`{url=}`", None

    # Check the nickname is not a URL
    if validators.url(nickname):
        return "It looks like the nickname you have entered is a URL.\n" \
                f"For security reasons, this is not allowed.\n`{nickname=}`", None


    feed_data, status_code = await get_rss_data(url)

    # Check the URL status code is valid
    if status_code != 200:
        return f"The URL provided returned an invalid status code:\n{url=}, {status_code=}", None

    # Check the contents is actually an RSS feed.
    feed = feedparser.parse(feed_data)
    if not feed.version:
        return f"The provided URL '{url}' does not seem to be a valid RSS feed.", None

    return None, feed


class CommandCog(commands.Cog):
    """
    Command cog.
    """

    def __init__(self, bot):
        super().__init__()
        self.bot = bot

    @commands.Cog.listener()
    async def on_ready(self):
        log.info(f"{self.__class__.__name__} cog is ready")

    async def source_autocomplete(self, inter: Interaction, nickname: str):
        """Provides RSS source autocomplete functionality for commands.

        Parameters
        ----------
        inter : Interaction
            Represents an app command interaction.
        nickname : str
            _description_

        Returns
        -------
        list of app_commands.Choice
            _description_
        """

        async with DatabaseManager() as database:
            whereclause = and_(
                RssSourceModel.discord_server_id == inter.guild_id,
                RssSourceModel.nick.ilike(f"%{nickname}%")
            )
            query = select(RssSourceModel).where(whereclause)
            result = await database.session.execute(query)
            sources = [
                Choice(name=rss.nick, value=rss.rss_url)
                for rss in result.scalars().all()
            ]

        return sources

    # All RSS commands belong to this group.
    rss_group = Group(
        name="rss",
        description="Commands for rss sources.",
        guild_only=True  # We store guild IDs in the database, so guild only = True
    )

    @rss_group.command(name="add")
    async def add_rss_source(self, inter: Interaction, nickname: str, url: str):
        """Add a new RSS source.

        Parameters
        ----------
        inter : Interaction
            Represents an app command interaction.
        nickname : str
            A name used to identify the RSS source.
        url : str
            The RSS feed URL.
        """

        await inter.response.defer()

        illegal_message, feed = await validate_rss_source(nickname, url)
        if illegal_message:
            await followup(inter, illegal_message, suppress_embeds=True)
            return

        log.debug("RSS feed added")

        async with DatabaseManager() as database:
            query = insert(RssSourceModel).values(
                discord_server_id = inter.guild_id,
                rss_url = url,
                nick=nickname
            )
            await database.session.execute(query)

            await audit(self,
                f"Added RSS source ({nickname=}, {url=})",
                inter.user.id, database=database
            )

        embed = Embed(title="RSS Feed Added", colour=Colour.dark_green())
        embed.add_field(name="Nickname", value=nickname)
        embed.add_field(name="URL", value=url)
        embed.set_thumbnail(url=feed.get("feed", {}).get("image", {}).get("href"))

        await followup(inter, embed=embed)

    @rss_group.command(name="remove")
    @autocomplete(url=source_autocomplete)
    async def remove_rss_source(self, inter: Interaction, url: str):
        """Delete an existing RSS source.

        Parameters
        ----------
        inter : Interaction
            Represents an app command interaction.
        url : str
            The RSS source to be removed. Autocomplete or enter the URL.
        """

        await inter.response.defer()

        log.debug(f"Attempting to remove RSS source ({url=})")

        async with DatabaseManager() as database:
            select_result = await database.session.execute(
                select(RssSourceModel).filter(
                    and_(
                        RssSourceModel.discord_server_id == inter.guild_id,
                        RssSourceModel.rss_url == url
                    )
                )
            )
            rss_source = select_result.scalars().one()
            nickname = rss_source.nick

            delete_result = await database.session.execute(
                delete(RssSourceModel).filter(
                    and_(
                        RssSourceModel.discord_server_id == inter.guild_id,
                        RssSourceModel.rss_url == url
                    )
                )
            )

            await audit(self,
                f"Added RSS source ({nickname=}, {url=})",
                inter.user.id, database=database
            )

        if not delete_result.rowcount:
            await followup(inter, "Couldn't find any RSS sources with this name.")
            return

        source = get_source(url)

        embed = Embed(title="RSS Feed Deleted", colour=Colour.dark_red())
        embed.add_field(name="Nickname", value=nickname)
        embed.add_field(name="URL", value=url)
        embed.set_thumbnail(url=source.icon_url)

        await followup(inter, embed=embed)

    @rss_group.command(name="list")
    @choices(sort=rss_list_sort_choices)
    async def list_rss_sources(self, inter: Interaction, sort: Choice[int]=None, sort_reverse: bool=False):
        """Provides a with a list of RSS sources available for the current server.

        Parameters
        ----------
        inter : Interaction
            Represents an app command interaction.
        """

        await inter.response.defer()

        # Default to the first choice if not specified.
        if type(sort) is Choice:
            description = "Sort by "
            description += "Nickname " if sort.value == 0 else "Date Added "
            description += '\U000025BC' if sort_reverse else '\U000025B2'
        else:
            sort = rss_list_sort_choices[0]
            description = ""

        sort = sort if type(sort) == Choice else rss_list_sort_choices[0]

        match sort.value, sort_reverse:
            case 0, False:
                order_by = RssSourceModel.nick.asc()
            case 0, True:
                order_by = RssSourceModel.nick.desc()
            case 1, False:                                # NOTE:
                order_by = RssSourceModel.created.desc()  # Datetime order is inversed because we want the latest
            case 1, True:                                 # date first, not the oldest as it would sort otherwise.
                order_by = RssSourceModel.created.asc()
            case _, _:
                raise ValueError("Unknown sort: %s" % sort)

        async with DatabaseManager() as database:
            whereclause = and_(RssSourceModel.discord_server_id == inter.guild_id)
            query = select(RssSourceModel).where(whereclause).order_by(order_by)
            result = await database.session.execute(query)

            rss_sources = result.scalars().all()

            if not rss_sources:
                await followup(inter, "It looks like you have no rss sources.")
                return

            output = "\n".join([f"{i}. **[{rss.nick}]({rss.rss_url})** " for i, rss in enumerate(rss_sources)])

        embed = Embed(
            title="Saved RSS Feeds",
            description=f"{description}\n\n{output}",
            colour=Colour.lighter_grey()
        )

        await followup(inter, embed=embed)

    @rss_group.command(name="fetch")
    @autocomplete(rss=source_autocomplete)
    async def fetch_rss(self, inter: Interaction, rss: str, max: int=1):
        # """"""

        await inter.response.defer()

        if max > 5:
            followup(inter, "It looks like you have requested too many articles.\nThe limit is 5")
            return

        invalid_message, feed = await validate_rss_source("", rss)
        if invalid_message:
            await followup(inter, invalid_message)
            return

        source = Source.from_parsed(feed)
        articles = source.get_latest_articles(max)

        embeds = []
        for article in articles:
            md_description = markdownify(article.description, strip=("img",))
            article_description = textwrap.shorten(md_description, 4096)

            embed = Embed(
                title=article.title,
                description=article_description,
                url=article.url,
                timestamp=article.published,
                colour=Colour.brand_red()
            )
            embed.set_thumbnail(url=source.icon_url)
            embed.set_image(url=await article.get_thumbnail_url())
            embed.set_footer(text=article.author)
            embed.set_author(
                name=source.name,
                url=source.url,
            )
            embeds.append(embed)

        async with DatabaseManager() as database:
            query = insert(SentArticleModel).values([
                {
                    "discord_server_id": inter.guild_id,
                    "discord_channel_id": inter.channel_id,
                    "discord_message_id": inter.id,
                    "article_url": article.url,
                }
                for article in articles
            ])
            await database.session.execute(query)
            await audit(self, f"User is requesting {max} articles from {source.name}", inter.user.id, database=database)

        await followup(inter, embeds=embeds)


async def setup(bot):
    """
    Setup function for this extension.
    Adds `CommandCog` to the bot.
    """

    cog = CommandCog(bot)
    await bot.add_cog(cog)
    log.info(f"Added {cog.__class__.__name__} cog")
